"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const index_1 = require("../../ContainerBase/index");
const checkParams_1 = require("../../../utils/checkParams");
const TreeNode_1 = __importDefault(require("./TreeNode"));
class TreeBaseContainer extends index_1.Container {
    constructor(cmp = (x, y) => {
        if (x < y)
            return -1;
        if (x > y)
            return 1;
        return 0;
    }) {
        super();
        this.root = undefined;
        this.header = new TreeNode_1.default();
        this._lowerBound = (curNode, key) => {
            if (curNode === undefined)
                return this.header;
            const cmpResult = this.cmp(curNode.key, key);
            if (cmpResult === 0)
                return curNode;
            if (cmpResult < 0)
                return this._lowerBound(curNode.right, key);
            const resNode = this._lowerBound(curNode.left, key);
            if (resNode === this.header)
                return curNode;
            return resNode;
        };
        this._upperBound = (curNode, key) => {
            if (curNode === undefined)
                return this.header;
            const cmpResult = this.cmp(curNode.key, key);
            if (cmpResult <= 0)
                return this._upperBound(curNode.right, key);
            const resNode = this._upperBound(curNode.left, key);
            if (resNode === this.header)
                return curNode;
            return resNode;
        };
        this._reverseLowerBound = (curNode, key) => {
            if (curNode === undefined)
                return this.header;
            const cmpResult = this.cmp(curNode.key, key);
            if (cmpResult === 0)
                return curNode;
            if (cmpResult > 0)
                return this._reverseLowerBound(curNode.left, key);
            const resNode = this._reverseLowerBound(curNode.right, key);
            if (resNode === this.header)
                return curNode;
            return resNode;
        };
        this._reverseUpperBound = (curNode, key) => {
            if (curNode === undefined)
                return this.header;
            const cmpResult = this.cmp(curNode.key, key);
            if (cmpResult >= 0)
                return this._reverseUpperBound(curNode.left, key);
            const resNode = this._reverseUpperBound(curNode.right, key);
            if (resNode === this.header)
                return curNode;
            return resNode;
        };
        this.inOrderTraversal = (curNode, callback) => {
            if (curNode === undefined)
                return false;
            const ifReturn = this.inOrderTraversal(curNode.left, callback);
            if (ifReturn)
                return true;
            if (callback(curNode))
                return true;
            return this.inOrderTraversal(curNode.right, callback);
        };
        this.cmp = cmp;
    }
    eraseNodeSelfBalance(curNode) {
        while (true) {
            const parentNode = curNode.parent;
            if (parentNode === this.header)
                return;
            if (curNode.color === TreeNode_1.default.red) {
                curNode.color = TreeNode_1.default.black;
                return;
            }
            if (curNode === parentNode.left) {
                const brother = parentNode.right;
                if (brother.color === TreeNode_1.default.red) {
                    brother.color = TreeNode_1.default.black;
                    parentNode.color = TreeNode_1.default.red;
                    if (parentNode === this.root) {
                        this.root = parentNode.rotateLeft();
                    }
                    else
                        parentNode.rotateLeft();
                }
                else if (brother.color === TreeNode_1.default.black) {
                    if (brother.right && brother.right.color === TreeNode_1.default.red) {
                        brother.color = parentNode.color;
                        parentNode.color = TreeNode_1.default.black;
                        brother.right.color = TreeNode_1.default.black;
                        if (parentNode === this.root) {
                            this.root = parentNode.rotateLeft();
                        }
                        else
                            parentNode.rotateLeft();
                        return;
                    }
                    else if (brother.left && brother.left.color === TreeNode_1.default.red) {
                        brother.color = TreeNode_1.default.red;
                        brother.left.color = TreeNode_1.default.black;
                        brother.rotateRight();
                    }
                    else {
                        brother.color = TreeNode_1.default.red;
                        curNode = parentNode;
                    }
                }
            }
            else {
                const brother = parentNode.left;
                if (brother.color === TreeNode_1.default.red) {
                    brother.color = TreeNode_1.default.black;
                    parentNode.color = TreeNode_1.default.red;
                    if (parentNode === this.root) {
                        this.root = parentNode.rotateRight();
                    }
                    else
                        parentNode.rotateRight();
                }
                else {
                    if (brother.left && brother.left.color === TreeNode_1.default.red) {
                        brother.color = parentNode.color;
                        parentNode.color = TreeNode_1.default.black;
                        brother.left.color = TreeNode_1.default.black;
                        if (parentNode === this.root) {
                            this.root = parentNode.rotateRight();
                        }
                        else
                            parentNode.rotateRight();
                        return;
                    }
                    else if (brother.right && brother.right.color === TreeNode_1.default.red) {
                        brother.color = TreeNode_1.default.red;
                        brother.right.color = TreeNode_1.default.black;
                        brother.rotateLeft();
                    }
                    else {
                        brother.color = TreeNode_1.default.red;
                        curNode = parentNode;
                    }
                }
            }
        }
    }
    eraseNode(curNode) {
        if (this.length === 1) {
            this.clear();
            return this.header;
        }
        let swapNode = curNode;
        while (swapNode.left || swapNode.right) {
            if (swapNode.right) {
                swapNode = swapNode.right;
                while (swapNode.left)
                    swapNode = swapNode.left;
            }
            else if (swapNode.left) {
                swapNode = swapNode.left;
                while (swapNode.right)
                    swapNode = swapNode.right;
            }
            [curNode.key, swapNode.key] = [swapNode.key, curNode.key];
            [curNode.value, swapNode.value] = [swapNode.value, curNode.value];
            curNode = swapNode;
        }
        if (this.header.left === swapNode) {
            this.header.left = swapNode.parent;
        }
        else if (this.header.right === swapNode) {
            this.header.right = swapNode.parent;
        }
        this.eraseNodeSelfBalance(swapNode);
        swapNode.remove();
        this.length -= 1;
        this.root.color = TreeNode_1.default.black;
    }
    insertNodeSelfBalance(curNode) {
        while (true) {
            const parentNode = curNode.parent;
            if (parentNode.color === TreeNode_1.default.black)
                return;
            const grandParent = parentNode.parent;
            if (parentNode === grandParent.left) {
                const uncle = grandParent.right;
                if (uncle && uncle.color === TreeNode_1.default.red) {
                    uncle.color = parentNode.color = TreeNode_1.default.black;
                    if (grandParent === this.root)
                        return;
                    grandParent.color = TreeNode_1.default.red;
                    curNode = grandParent;
                    continue;
                }
                else if (curNode === parentNode.right) {
                    parentNode.rotateLeft();
                    curNode.color = TreeNode_1.default.black;
                }
                else
                    parentNode.color = TreeNode_1.default.black;
                grandParent.color = TreeNode_1.default.red;
                if (grandParent === this.root) {
                    this.root = grandParent.rotateRight();
                }
                else
                    grandParent.rotateRight();
            }
            else {
                const uncle = grandParent.left;
                if (uncle && uncle.color === TreeNode_1.default.red) {
                    uncle.color = parentNode.color = TreeNode_1.default.black;
                    if (grandParent === this.root)
                        return;
                    grandParent.color = TreeNode_1.default.red;
                    curNode = grandParent;
                    continue;
                }
                else if (curNode === parentNode.left) {
                    parentNode.rotateRight();
                    curNode.color = TreeNode_1.default.black;
                }
                else
                    parentNode.color = TreeNode_1.default.black;
                grandParent.color = TreeNode_1.default.red;
                if (grandParent === this.root) {
                    this.root = grandParent.rotateLeft();
                }
                else
                    grandParent.rotateLeft();
            }
            return;
        }
    }
    findElementNode(curNode, key) {
        while (curNode) {
            const cmpResult = this.cmp(curNode.key, key);
            if (cmpResult < 0) {
                curNode = curNode.right;
            }
            else if (cmpResult > 0) {
                curNode = curNode.left;
            }
            else
                return curNode;
        }
        return curNode;
    }
    set(key, value) {
        if (this.root === undefined) {
            this.length += 1;
            this.root = new TreeNode_1.default(key, value);
            this.root.color = TreeNode_1.default.black;
            this.root.parent = this.header;
            this.header.parent = this.root;
            this.header.left = this.root;
            this.header.right = this.root;
            return;
        }
        let curNode;
        const minNode = this.header.left;
        const compareToMin = this.cmp(minNode.key, key);
        if (compareToMin === 0) {
            minNode.value = value;
            return;
        }
        else if (compareToMin > 0) {
            minNode.left = new TreeNode_1.default(key, value);
            minNode.left.parent = minNode;
            curNode = minNode.left;
            this.header.left = curNode;
        }
        else {
            const maxNode = this.header.right;
            const compareToMax = this.cmp(maxNode.key, key);
            if (compareToMax === 0) {
                maxNode.value = value;
                return;
            }
            else if (compareToMax < 0) {
                maxNode.right = new TreeNode_1.default(key, value);
                maxNode.right.parent = maxNode;
                curNode = maxNode.right;
                this.header.right = curNode;
            }
            else {
                curNode = this.root;
                while (true) {
                    const cmpResult = this.cmp(curNode.key, key);
                    if (cmpResult > 0) {
                        if (curNode.left === undefined) {
                            curNode.left = new TreeNode_1.default(key, value);
                            curNode.left.parent = curNode;
                            curNode = curNode.left;
                            break;
                        }
                        curNode = curNode.left;
                    }
                    else if (cmpResult < 0) {
                        if (curNode.right === undefined) {
                            curNode.right = new TreeNode_1.default(key, value);
                            curNode.right.parent = curNode;
                            curNode = curNode.right;
                            break;
                        }
                        curNode = curNode.right;
                    }
                    else {
                        curNode.value = value;
                        return;
                    }
                }
            }
        }
        this.length += 1;
        this.insertNodeSelfBalance(curNode);
    }
    clear() {
        this.length = 0;
        this.root = undefined;
        this.header.parent = undefined;
        this.header.left = this.header.right = undefined;
    }
    eraseElementByPos(pos) {
        (0, checkParams_1.checkWithinAccessParams)(pos, 0, this.length - 1);
        let index = 0;
        this.inOrderTraversal(this.root, curNode => {
            if (pos === index) {
                this.eraseNode(curNode);
                return true;
            }
            index += 1;
            return false;
        });
    }
    /**
     * Removes the elements of the specified key.
     */
    eraseElementByKey(key) {
        if (!this.length)
            return;
        const curNode = this.findElementNode(this.root, key);
        if (curNode === undefined)
            return;
        this.eraseNode(curNode);
    }
    /**
     * @return An iterator point to the next iterator.
     * Removes element by iterator.
     */
    eraseElementByIterator(iter) {
        // @ts-ignore
        const node = iter.node;
        if (node === this.header) {
            throw new RangeError('Invalid iterator');
        }
        if (node.right === undefined) {
            iter = iter.next();
        }
        this.eraseNode(node);
        return iter;
    }
    /**
     * @return The height of the RB-tree.
     */
    getHeight() {
        if (!this.length)
            return 0;
        const traversal = (curNode) => {
            if (!curNode)
                return 1;
            return Math.max(traversal(curNode.left), traversal(curNode.right)) + 1;
        };
        return traversal(this.root);
    }
}
exports.default = TreeBaseContainer;
